"""
Our DC sources ultimately need to follow this standard form if we want to enable full electrical integration.

```
11.3.2 .DC: DC Transfer Function

General form:

    .dc srcnam vstart vstop vincr [src2 start2 stop2 incr2]

Examples:

    .dc VIN 0.25 5.0 0.25
    .dc VDS 0 10 .5 VGS 0 5 1
    .dc VCE 0 10 .25 IB 0 10u 1u
    .dc RLoad 1k 2k 100
    .dc TEMP -15 75 5
```

"""

from __future__ import annotations

from typing import Literal, Optional

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.types import ArrayFloat1D
from tidy3d.constants import AMP, VOLT
from tidy3d.constants import inf as td_inf


class DCVoltageSource(Tidy3dBaseModel):
    """
    DC voltage source in volts.

    Notes
    -----

        This voltage refers to potential above the equivalent simulation ground. Currently, electrical ports
        are not defined.

    Examples
    --------
    >>> import tidy3d as td
    >>> voltages = [-0.5, 0, 1, 2, 3, 4]
    >>> voltage_source = td.DCVoltageSource(voltage=voltages)
    """

    name: Optional[str]
    voltage: ArrayFloat1D = pd.Field(
        ...,
        title="Voltage",
        description="DC voltage usually used as source in 'VoltageBC' boundary conditions.",
        units=VOLT,
    )

    # TODO: This should have always been in the field above but was introduced wrongly as a
    # standalone field. Keeping for compatibility, remove in 3.0.
    units: Literal[VOLT] = VOLT

    @pd.validator("voltage")
    def check_voltage(cls, val):
        for v in val:
            if v == td_inf:
                raise ValueError(f"Voltages must be finite. Currently  voltage={val}.")
        return val


class DCCurrentSource(Tidy3dBaseModel):
    """
    DC current source in amperes.

    Example
    -------
    >>> import tidy3d as td
    >>> current_source = td.DCCurrentSource(current=0.4)
    """

    name: Optional[str]
    current: pd.FiniteFloat = pd.Field(
        title="Current",
        description="DC current usually used as source in 'CurrentBC' boundary conditions.",
        units=AMP,
    )

    # TODO: This should have always been in the field above but was introduced wrongly as a
    # standalone field. Keeping for compatibility, remove in 3.0.
    units: Literal[AMP] = AMP



"""Defines heat material specifications"""

from __future__ import annotations

import pydantic.v1 as pd

from tidy3d.components.tcad.boundary.abstract import HeatChargeBC
from tidy3d.constants import HEAT_FLUX, HEAT_TRANSFER_COEFF, KELVIN


class TemperatureBC(HeatChargeBC):
    """Constant temperature thermal boundary conditions.

    Example
    -------
    >>> import tidy3d as td
    >>> bc = td.TemperatureBC(temperature=300)
    """

    temperature: pd.PositiveFloat = pd.Field(
        title="Temperature",
        description=f"Temperature value in units of {KELVIN}.",
        units=KELVIN,
    )


class HeatFluxBC(HeatChargeBC):
    """Constant flux thermal boundary conditions.

    Example
    -------
    >>> import tidy3d as td
    >>> bc = td.HeatFluxBC(flux=1)
    """

    flux: float = pd.Field(
        title="Heat Flux",
        description=f"Heat flux value in units of {HEAT_FLUX}.",
        units=HEAT_FLUX,
    )


class ConvectionBC(HeatChargeBC):
    """Convective thermal boundary conditions.

    Example
    -------
    >>> import tidy3d as td
    >>> bc = td.ConvectionBC(ambient_temperature=300, transfer_coeff=1)
    """

    ambient_temperature: pd.PositiveFloat = pd.Field(
        title="Ambient Temperature",
        description=f"Ambient temperature value in units of {KELVIN}.",
        units=KELVIN,
    )

    transfer_coeff: pd.NonNegativeFloat = pd.Field(
        title="Heat Transfer Coefficient",
        description=f"Heat flux value in units of {HEAT_TRANSFER_COEFF}.",
        units=HEAT_TRANSFER_COEFF,
    )



"""Imports for adjoint plugin."""
# ruff: noqa: E402

# import the jax version of tidy3d components
from __future__ import annotations

from textwrap import dedent

from tidy3d.log import log

_DOC_URL = "https://github.com/flexcompute/tidy3d/blob/develop/tidy3d/plugins/autograd/README.md"

_MSG = dedent(
    f"""
    The 'adjoint' plugin (legacy JAX-based adjoint plugin) was deprecated in Tidy3D '2.7.0' and will be removed in '2.10.0'.

    Migrate to the native autograd workflow:
        import tidy3d as td
        import autograd.numpy as np
        from autograd import grad

    It uses standard 'td.' objects, has fewer dependencies, and offers a smoother optimization experience.
    Full guide: {_DOC_URL}
    """
).strip()

log.warning(_MSG)

try:
    import jax

    jax.config.update("jax_enable_x64", True)
except ImportError as e:
    raise ImportError(
        "The 'jax' package is required for adjoint plugin. We were not able to import it. "
        "To get the appropriate packages for your system, install tidy3d using '[jax]' option, "
        "for example: $pip install 'tidy3d[jax]'."
    ) from e

from .components.data.data_array import JaxDataArray
from .components.data.dataset import JaxPermittivityDataset
from .components.data.monitor_data import JaxModeData
from .components.data.sim_data import JaxSimulationData
from .components.geometry import JaxBox, JaxComplexPolySlab, JaxGeometryGroup, JaxPolySlab
from .components.medium import JaxAnisotropicMedium, JaxCustomMedium, JaxMedium
from .components.simulation import JaxSimulation
from .components.structure import (
    JaxStructure,
    JaxStructureStaticGeometry,
    JaxStructureStaticMedium,
)
from .web import run, run_async

__all__ = [
    "JaxAnisotropicMedium",
    "JaxBox",
    "JaxComplexPolySlab",
    "JaxCustomMedium",
    "JaxDataArray",
    "JaxGeometryGroup",
    "JaxMedium",
    "JaxModeData",
    "JaxPermittivityDataset",
    "JaxPolySlab",
    "JaxSimulation",
    "JaxSimulationData",
    "JaxStructure",
    "JaxStructureStaticGeometry",
    "JaxStructureStaticMedium",
    "run",
    "run_async",
]



"""Objects that define how data is recorded from simulation."""

from __future__ import annotations

from math import isclose
from typing import Literal

import pydantic.v1 as pd

from tidy3d.components.tcad.monitors.abstract import HeatChargeMonitor


class VolumeMeshMonitor(HeatChargeMonitor):
    """Monitor recording the volume mesh. The monitor size must be either 2D or 3D. If a 2D monitor
    is used in a 3D simulation, the sliced volumetric mesh on the plane of the monitor will be
    stored as a ``TriangularGridDataset``."""

    unstructured: Literal[True] = pd.Field(
        True,
        title="Unstructured Grid",
        description="Return the original unstructured grid.",
    )

    @pd.validator("size", always=True)
    def _at_least_2d(cls, val):
        """Validate that the monitor has at least two non-zero dimensions."""
        if len([d for d in val if isclose(d, 0)]) > 1:
            raise ValueError("'VolumeMeshMonitor' must have at least two nonzero dimensions.")
        return val



# backwards compatibility support for ``from tidy3d.plugins.smatrix.smatrix import ``
from __future__ import annotations

from .component_modelers.modal import ComponentModeler
from .ports.modal import Port

__all__ = ["ComponentModeler", "Port"]



from __future__ import annotations

from .filters import (
    CircularFilter,
    ConicFilter,
    make_circular_filter,
    make_conic_filter,
    make_filter,
)
from .misc import grey_indicator
from .parametrizations import FilterAndProject, make_filter_and_project
from .penalties import ErosionDilationPenalty, make_curvature_penalty, make_erosion_dilation_penalty
from .projections import ramp_projection, tanh_projection

__all__ = [
    "CircularFilter",
    "ConicFilter",
    "ErosionDilationPenalty",
    "FilterAndProject",
    "grey_indicator",
    "make_circular_filter",
    "make_conic_filter",
    "make_curvature_penalty",
    "make_erosion_dilation_penalty",
    "make_filter",
    "make_filter_and_project",
    "ramp_projection",
    "tanh_projection",
]



from __future__ import annotations

from typing import Optional

import numpy as np
from autograd.extend import defvjp, primitive
from numpy.typing import NDArray

from tidy3d.log import log


def _assert_strictly_monotonic(x: NDArray) -> None:
    """Raise if ``x`` is not strictly monotonic (all increasing or all decreasing)."""
    dx = np.diff(x)
    if not (np.all(dx > 0) or np.all(dx < 0)):
        raise ValueError(
            "'x' must be strictly monotonic; mixed or duplicated knots "
            "cause division‑by‑zero or wrong interval selection."
        )


def get_interval_indices(x: NDArray, x_eval: NDArray) -> NDArray:
    """Return the index of the interval in which each ``x_eval`` value lies.

    Parameters
    ----------
    x : np.ndarray
        Array of points defining intervals
    x_eval : np.ndarray
        Points to find intervals for

    Returns
    -------
    np.ndarray
        Indices of intervals
    """
    # use right‑continuity to avoid a left‑side step when x_eval == x[i]
    idx = np.searchsorted(x[:-1], x_eval, side="right") - 1
    return np.clip(idx, 0, len(x) - 2)


def accumulate_polynomial_factors(
    x_points: NDArray, x_eval: NDArray, g: NDArray, order: int
) -> tuple[NDArray, NDArray]:
    """Given an array of evaluation points ``x_eval`` and the corresponding gradient
    contributions ``g``, compute the polynomial factors accumulated per interval.

    Parameters
    ----------
    x_points : np.ndarray
        Points defining the intervals
    x_eval : np.ndarray
        Evaluation points
    g : np.ndarray
        Gradient contributions
    order : int
        Polynomial order

    Returns
    -------
    poly_factors : np.ndarray
        Shape ``(n, order+1)``, where ``n = len(x_points)-1``.
        ``poly_factors[i, k]`` holds the sum of ``g_j * (dx_j**k)`` for all
        ``x_eval_j`` in the ``i``-th interval.
    idx : np.ndarray
        Indices of intervals for each ``x_eval`` point.
    """
    n = len(x_points) - 1
    idx = get_interval_indices(x_points, x_eval)
    dx = x_eval - x_points[idx]

    # (m, order+1)
    powers = dx[:, None] ** np.arange(order + 1)
    poly_factors = np.zeros((n, order + 1))
    np.add.at(poly_factors, idx, g[:, None] * powers)

    return poly_factors, idx


def compute_linear_coefficients(x: NDArray, y: NDArray) -> tuple[NDArray, NDArray]:
    """Compute linear spline coefficients.

    .. math::
        S_i(t) = a[i] + b[i](t - x[i])

    Parameters
    ----------
    x : np.ndarray
        X coordinates
    y : np.ndarray
        Y coordinates

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Coefficients ``(a, b)``
    """
    a = y[:-1].copy()
    b = (y[1:] - y[:-1]) / (x[1:] - x[:-1])
    return a, b


def evaluate_linear_spline(
    x: NDArray,
    coeffs: tuple[NDArray, NDArray],
    x_eval: NDArray,
) -> NDArray:
    """Evaluate a linear spline at the specified points.

    Parameters
    ----------
    x : np.ndarray
        X coordinates defining the spline
    coeffs : tuple[np.ndarray, np.ndarray]
        Spline coefficients ``(a, b)``
    x_eval : np.ndarray
        Points at which to evaluate the spline

    Returns
    -------
    np.ndarray
        Evaluated spline values
    """
    a, b = coeffs
    idx = get_interval_indices(x, x_eval)
    dx = x_eval - x[idx]
    return a[idx] + b[idx] * dx


def get_linear_derivative_wrt_y(
    x: NDArray,
    y: NDArray,
) -> tuple[NDArray, NDArray]:
    """Compute derivative of linear spline coefficients wrt ``y``.

    Parameters
    ----------
    x : np.ndarray
        X coordinates
    y : np.ndarray
        Y coordinates

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        ``(da_dy, db_dy)``
    """
    n = len(x) - 1
    da_dy = np.zeros((n, len(y)))
    db_dy = np.zeros((n, len(y)))

    for i in range(n):
        # a[i] = y[i]
        da_dy[i, i] = 1.0

        # b[i] = (y[i+1] - y[i]) / (x[i+1] - x[i])
        h_i = x[i + 1] - x[i]
        db_dy[i, i] = -1.0 / h_i
        db_dy[i, i + 1] = 1.0 / h_i

    return da_dy, db_dy


def compute_quadratic_coefficients(
    x: NDArray,
    y: NDArray,
    left_deriv: Optional[float] = None,
) -> tuple[NDArray, NDArray, NDArray]:
    """Compute quadratic spline coefficients.

    .. math::
        S_i(t) = a[i] + b[i]*(t - x[i]) + c[i]*(t - x[i])^2

    Parameters
    ----------
    x : np.ndarray
        X coordinates
    y : np.ndarray
        Y coordinates
    left_deriv : float = None
        Left endpoint derivative

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        Coefficients ``(a, b, c)``
    """
    n = len(x) - 1
    h = np.diff(x)

    a = y[:-1].copy()
    b = np.zeros(n)
    c = np.zeros(n)

    # If left derivative is provided, use it for the first segment
    if left_deriv is not None:
        b[0] = left_deriv
    else:
        b[0] = (y[1] - y[0]) / h[0]

    c[0] = (y[1] - y[0] - b[0] * h[0]) / (h[0] ** 2)

    # For each internal point, ensure continuity of function + 1st derivative
    for i in range(1, n):
        deriv_end = b[i - 1] + 2 * c[i - 1] * h[i - 1]
        b[i] = deriv_end
        c[i] = (y[i + 1] - y[i] - b[i] * h[i]) / (h[i] ** 2)

    return a, b, c


def evaluate_quadratic_spline(
    x: NDArray,
    coeffs: tuple[NDArray, NDArray, NDArray],
    x_eval: NDArray,
) -> NDArray:
    """Evaluate a quadratic spline at the specified points.

    Parameters
    ----------
    x : np.ndarray
        X coordinates defining the spline
    coeffs : tuple[np.ndarray, np.ndarray, np.ndarray]
        Spline coefficients ``(a, b, c)``
    x_eval : np.ndarray
        Points at which to evaluate the spline

    Returns
    -------
    np.ndarray
        Evaluated spline values
    """
    a, b, c = coeffs
    idx = get_interval_indices(x, x_eval)
    dx = x_eval - x[idx]
    return a[idx] + b[idx] * dx + c[idx] * dx**2


def get_quadratic_derivative_wrt_y(
    x: NDArray,
    y: NDArray,
    left_deriv: Optional[float] = None,
) -> tuple[NDArray, NDArray, NDArray]:
    """Compute derivative of quadratic spline coefficients wrt ``y``.

    Parameters
    ----------
    x : np.ndarray
        X coordinates
    y : np.ndarray
        Y coordinates
    left_deriv : float = None
        Left endpoint derivative

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        ``(da_dy, db_dy, dc_dy)``
    """
    n = len(x) - 1
    h = np.diff(x)

    da_dy = np.zeros((n, len(y)))
    db_dy = np.zeros((n, len(y)))
    dc_dy = np.zeros((n, len(y)))

    for i in range(n):
        da_dy[i, i] = 1.0

    if left_deriv is None:
        db_dy[0, 0] = -1.0 / h[0]
        db_dy[0, 1] = 1.0 / h[0]
        dc_dy[0, :] = 0.0
    else:
        dc_dy[0, 0] = -1.0 / (h[0] ** 2)
        dc_dy[0, 1] = 1.0 / (h[0] ** 2)

    for i in range(1, n):
        # Derivative wrt y_j: db_dy[i, j] = db_dy[i-1, j] + 2*h[i-1]*dc_dy[i-1, j]
        db_dy[i, :] = db_dy[i - 1, :] + 2.0 * h[i - 1] * dc_dy[i - 1, :]

        # First term: -db_dy[i,j] / h[i]
        dc_dy[i, :] = -(db_dy[i, :] * h[i]) / (h[i] ** 2)

        # Second term: Add direct contributions from y[i] and y[i+1]
        dc_dy[i, i] += -1.0 / (h[i] ** 2)
        dc_dy[i, i + 1] += 1.0 / (h[i] ** 2)

    return da_dy, db_dy, dc_dy


def setup_cubic_tridiagonal(
    x: NDArray,
    y: NDArray,
    endpoint_derivs: tuple[Optional[float], Optional[float]] = (None, None),
) -> tuple[NDArray, NDArray, NDArray, NDArray, NDArray]:
    """Return (lower, diag, upper, rhs, h) for the cubic spline system.

    Parameters
    ----------
    x : np.ndarray
        X coordinates
    y : np.ndarray
        Y coordinates
    endpoint_derivs : tuple[float, float] = (None, None)
        Derivatives at endpoints (left, right)

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        ``(lower, diag, upper, rhs, h)`` for the tridiagonal system
    """
    n = len(x) - 1
    h = np.diff(x)

    lower = np.zeros(n + 1)
    diag = np.zeros(n + 1)
    upper = np.zeros(n + 1)
    rhs = np.zeros(n + 1)

    left_deriv, right_deriv = endpoint_derivs

    for i in range(1, n):
        lower[i] = h[i - 1]
        diag[i] = 2.0 * (h[i - 1] + h[i])
        upper[i] = h[i]
        rhs[i] = (3.0 / h[i]) * (y[i + 1] - y[i]) - (3.0 / h[i - 1]) * (y[i] - y[i - 1])

    # Left boundary
    if left_deriv is None:
        diag[0] = 2.0 * h[0]
        upper[0] = h[0]
        rhs[0] = 3.0 * (y[1] - y[0]) / h[0]
    else:
        diag[0] = 2.0
        upper[0] = 1.0
        rhs[0] = (3.0 / h[0]) * ((y[1] - y[0]) / h[0] - left_deriv)

    # Right boundary
    if right_deriv is None:
        lower[n] = h[n - 1]
        diag[n] = 2.0 * h[n - 1]
        rhs[n] = 3.0 * (y[n] - y[n - 1]) / h[n - 1]
    else:
        lower[n] = 1.0
        diag[n] = 2.0
        rhs[n] = (3.0 / h[n - 1]) * (right_deriv - (y[n] - y[n - 1]) / h[n - 1])

    return lower, diag, upper, rhs, h


def _solve_tridiagonal(lower: NDArray, diag: NDArray, upper: NDArray, rhs: NDArray) -> NDArray:
    """Solve tridiagonal system (single RHS) using the Thomas algorithm.

    Parameters
    ----------
    lower : np.ndarray
        Lower diagonal elements
    diag : np.ndarray
        Main diagonal elements
    upper : np.ndarray
        Upper diagonal elements
    rhs : np.ndarray
        Right-hand side vector

    Returns
    -------
    np.ndarray
        Solution vector
    """
    from scipy.linalg import solve_banded

    n = diag.size
    ab = np.zeros((3, n))
    ab[0, 1:] = upper[:-1]
    ab[1, :] = diag
    ab[2, :-1] = lower[1:]
    return solve_banded((1, 1), ab, rhs, overwrite_ab=False, overwrite_b=False, check_finite=False)


def _solve_tridiagonal_multi(lower: NDArray, diag: NDArray, upper: NDArray, B: NDArray) -> NDArray:
    """Solve tridiagonal system with multiple RHS (columns of *B*).

    Parameters
    ----------
    lower : np.ndarray
        Lower diagonal elements
    diag : np.ndarray
        Main diagonal elements
    upper : np.ndarray
        Upper diagonal elements
    B : np.ndarray
        Right-hand side matrix

    Returns
    -------
    np.ndarray
        Solution matrix
    """
    n, k = B.shape
    c = upper.copy()
    d = diag.copy()
    a = lower.copy()
    W = B.copy()

    for i in range(1, n):
        m = a[i] / d[i - 1]
        d[i] -= m * c[i - 1]
        W[i, :] -= m * W[i - 1, :]

    X = np.empty_like(W)
    X[-1, :] = W[-1, :] / d[-1]
    for i in range(n - 2, -1, -1):
        X[i, :] = (W[i, :] - c[i] * X[i + 1, :]) / d[i]
    return X


def compute_cubic_coefficients(
    x: NDArray,
    y: NDArray,
    c_full: NDArray,
    h: NDArray,
) -> tuple[NDArray, NDArray, NDArray, NDArray]:
    """Computes cubic spline coefficients from the solved ``c`` array:

    .. math::
        S_i(t) = a[i] + b[i](t - x[i]) + c[i](t - x[i])^2 + d[i](t - x[i])^3.

    Parameters
    ----------
    x : np.ndarray
        X coordinates
    y : np.ndarray
        Y coordinates
    c_full : np.ndarray
        Solved ``c`` values
    h : np.ndarray
        Interval widths

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        Coefficients ``(a, b, c, d)``
    """
    n = len(x) - 1
    a = y[:-1].copy()
    b = np.zeros(n)
    d = np.zeros(n)

    for i in range(n):
        b[i] = (y[i + 1] - y[i]) / h[i] - (h[i] / 3.0) * (2 * c_full[i] + c_full[i + 1])
        d[i] = (c_full[i + 1] - c_full[i]) / (3.0 * h[i])

    return a, b, c_full[:-1], d


def evaluate_cubic_spline(
    x: NDArray,
    coeffs: tuple[NDArray, NDArray, NDArray, NDArray],
    x_eval: NDArray,
) -> NDArray:
    """Evaluate a cubic spline at the specified points.

    Parameters
    ----------
    x : np.ndarray
        X coordinates defining the spline
    coeffs : tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        Spline coefficients (a, b, c, d)
    x_eval : np.ndarray
        Points at which to evaluate the spline

    Returns
    -------
    np.ndarray
        Evaluated spline values
    """
    a, b, c, d = coeffs
    idx = get_interval_indices(x, x_eval)
    dx = x_eval - x[idx]
    return a[idx] + b[idx] * dx + c[idx] * dx**2 + d[idx] * dx**3


def compute_spline_coefficients(
    x: NDArray,
    y: NDArray,
    endpoint_derivs: tuple[Optional[float], Optional[float]] = (None, None),
) -> tuple[NDArray, NDArray, NDArray, NDArray]:
    """Compute the cubic spline coefficients ``(a, b, c, d)``.

    Parameters
    ----------
    x : np.ndarray
        X coordinates
    y : np.ndarray
        Y coordinates
    endpoint_derivs : tuple[float, float] = (None, None)
        Derivatives at endpoints (left, right)

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        Coefficients ``(a, b, c, d)``
    """
    lower, diag, upper, rhs, h = setup_cubic_tridiagonal(x, y, endpoint_derivs)
    c_full = _solve_tridiagonal(lower, diag, upper, rhs)
    return compute_cubic_coefficients(x, y, c_full, h)


def get_cubic_derivative_wrt_y(
    x: NDArray,
    y: NDArray,
    endpoint_derivs: tuple[Optional[float], Optional[float]] = (None, None),
) -> tuple[NDArray, NDArray, NDArray, NDArray, NDArray]:
    """Compute derivatives of cubic spline coefficients ``(a, b, c, d)``
    wrt ``y`` values.

    Parameters
    ----------
    x : np.ndarray
        X coordinates
    y : np.ndarray
        Y coordinates
    endpoint_derivs : tuple[float, float] = (None, None)
        Derivatives at endpoints (left, right)

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        ``(da_dy, db_dy, dc_dy, dd_dy, h)``
    """
    n = len(x) - 1
    lower, diag, upper, _, h = setup_cubic_tridiagonal(x, y, endpoint_derivs)

    db_dy = np.zeros((n + 1, len(y)))
    left_deriv, right_deriv = endpoint_derivs

    for i in range(1, n):
        db_dy[i, i + 1] += 3.0 / h[i]
        db_dy[i, i] += -3.0 / h[i] - 3.0 / h[i - 1]
        db_dy[i, i - 1] += 3.0 / h[i - 1]

    if left_deriv is None:
        db_dy[0, 0] = -3.0 / h[0]
        db_dy[0, 1] = 3.0 / h[0]
    else:
        db_dy[0, 0] = (3.0 / h[0]) * (-1.0 / h[0])
        db_dy[0, 1] = (3.0 / h[0]) * (+1.0 / h[0])

    if right_deriv is None:
        db_dy[n, n] = 3.0 / h[n - 1]
        db_dy[n, n - 1] = -3.0 / h[n - 1]
    else:
        db_dy[n, n] = (3.0 / h[n - 1]) * (-1.0 / h[n - 1])
        db_dy[n, n - 1] = (3.0 / h[n - 1]) * (+1.0 / h[n - 1])

    dc_dy = _solve_tridiagonal_multi(lower, diag, upper, db_dy)  # shape (n+1, len(y))

    da_dy = np.zeros((n, len(y)))
    db_dy_out = np.zeros((n, len(y)))
    dd_dy = np.zeros((n, len(y)))

    for i in range(n):
        da_dy[i, i] = 1.0

        # b[i] depends on y[i], y[i+1], and c[i], c[i+1]
        # b[i] = (y[i+1]-y[i])/h[i] - h[i]/3 * (2*c[i] + c[i+1])
        # => partial wrt y[i+1] => +1/h[i]
        # => partial wrt y[i]   => -1/h[i]
        db_dy_out[i, i] += -1.0 / h[i]
        if i + 1 < len(y):
            db_dy_out[i, i + 1] += 1.0 / h[i]

        # partial wrt c[i], c[i+1]
        # => - h[i]/3 * (2 * dc_dy[i, :] + dc_dy[i+1, :])
        db_dy_out[i, :] += -(h[i] / 3.0) * (2.0 * dc_dy[i, :] + dc_dy[i + 1, :])

        # d[i] = (c[i+1] - c[i]) / (3*h[i])
        dd_dy[i, :] = (dc_dy[i + 1, :] - dc_dy[i, :]) / (3.0 * h[i])

    return da_dy, db_dy_out, dc_dy[:-1], dd_dy, h


def compute_spline_coeffs(
    x_points: NDArray,
    y_points: NDArray,
    endpoint_derivatives: tuple[Optional[float], Optional[float]] = (None, None),
    order: int = 3,
) -> tuple:
    """Compute spline coefficients for the given order.

    Parameters
    ----------
    x_points : np.ndarray
        X coordinates
    y_points : np.ndarray
        Y coordinates
    endpoint_derivatives : tuple[float, float] = (None, None)
        Derivatives at endpoints (left, right)
    order : int = 3
        Spline order

    Returns
    -------
    tuple
        order=1 => ``(a, b)``
        order=2 => ``(a, b, c)``
        order=3 => ``(a, b, c, d)``
    """
    if order == 1:
        return compute_linear_coefficients(x_points, y_points)
    if order == 2:
        left_deriv = endpoint_derivatives[0]
        return compute_quadratic_coefficients(x_points, y_points, left_deriv)
    if order == 3:
        return compute_spline_coefficients(x_points, y_points, endpoint_derivatives)
    raise NotImplementedError(f"Spline order '{order}' not implemented.")


def evaluate_spline(x_points: NDArray, coeffs: tuple, x_eval: NDArray) -> NDArray:
    """Evaluate a spline at the specified points.

    Parameters
    ----------
    x_points : np.ndarray
        X coordinates defining the spline
    coeffs : tuple
        Spline coefficients
    x_eval : np.ndarray
        Points at which to evaluate the spline

    Returns
    -------
    np.ndarray
        Evaluated spline values
    """
    order = len(coeffs) - 1

    if order == 1:
        return evaluate_linear_spline(x_points, coeffs, x_eval)
    if order == 2:
        return evaluate_quadratic_spline(x_points, coeffs, x_eval)
    if order == 3:
        return evaluate_cubic_spline(x_points, coeffs, x_eval)
    raise NotImplementedError(f"Spline order '{order}' not implemented.")


def get_spline_derivatives_wrt_y(
    order: int,
    x_points: NDArray,
    y_points: NDArray,
    endpoint_derivatives: tuple[Optional[float], Optional[float]] = (None, None),
):
    """Returns a tuple of derivative arrays for the given spline order.

    Parameters
    ----------
    order : int
        Spline order
    x_points : np.ndarray
        X coordinates
    y_points : np.ndarray
        Y coordinates
    endpoint_derivatives : tuple[float, float] = (None, None)
        Derivatives at endpoints (left, right)

    Returns
    -------
    tuple
        order=1 => ``(da_dy, db_dy)``
        order=2 => ``(da_dy, db_dy, dc_dy)``
        order=3 => ``(da_dy, db_dy, dc_dy, dd_dy, h)``
    """
    if order == 1:
        return get_linear_derivative_wrt_y(x_points, y_points)
    if order == 2:
        left_deriv = endpoint_derivatives[0]
        return get_quadratic_derivative_wrt_y(x_points, y_points, left_deriv)
    if order == 3:
        return get_cubic_derivative_wrt_y(x_points, y_points, endpoint_derivatives)
    raise NotImplementedError(f"Derivatives for spline order '{order}' not implemented.")


@primitive
def _interpolate_spline(
    x_points: NDArray,
    y_points: NDArray,
    num_points: int,
    order: int,
    endpoint_derivatives: tuple[Optional[float], Optional[float]] = (None, None),
) -> tuple[NDArray, NDArray]:
    """Primitive function to perform spline interpolation of a given order
    with optional endpoint derivatives.

    Autograd requires that arguments to primitives are passed in positionally.
    ``interpolate_spline`` is the public-facing wrapper for this function,
    which allows keyword arguments in case users pass in kwargs.
    """
    if order not in (1, 2, 3):
        raise NotImplementedError(f"Spline order '{order}' not implemented.")

    if x_points.shape != y_points.shape:
        raise ValueError("'x_points' and 'y_points' must have the same shape")

    if order == 1 and (endpoint_derivatives[0] is not None or endpoint_derivatives[1] is not None):
        log.warning("Endpoint derivatives are ignored for linear splines ('order=1').")
    elif order == 2 and endpoint_derivatives[1] is not None:
        log.warning("Right endpoint derivative is ignored for quadratic splines ('order=2').")

    _assert_strictly_monotonic(x_points)

    # allow decreasing input by reversing internally
    reversed_order = x_points[0] > x_points[-1]
    if reversed_order:
        x_points_proc = x_points[::-1]
        y_points_proc = y_points[::-1]
        endpoint_derivatives_proc = (endpoint_derivatives[1], endpoint_derivatives[0])
    else:
        x_points_proc = x_points
        y_points_proc = y_points
        endpoint_derivatives_proc = endpoint_derivatives

    x_min, x_max = x_points_proc[0], x_points_proc[-1]
    x_eval = np.linspace(x_min, x_max, num_points)

    coeffs = compute_spline_coeffs(x_points_proc, y_points_proc, endpoint_derivatives_proc, order)
    y_eval = evaluate_spline(x_points_proc, coeffs, x_eval)

    if reversed_order:
        return x_eval[::-1], y_eval[::-1]
    return x_eval, y_eval


def interpolate_spline_y_vjp(ans, x_points, y_points, num_points, order, endpoint_derivatives):
    """VJP for interpolate_spline wrt y_points."""

    def vjp(g):
        reversed_order = x_points[0] > x_points[-1]
        if reversed_order:
            x_proc = x_points[::-1]
            y_proc = y_points[::-1]
            endpoint_derivatives_proc = (endpoint_derivatives[1], endpoint_derivatives[0])
        else:
            x_proc = x_points
            y_proc = y_points
            endpoint_derivatives_proc = endpoint_derivatives

        derivative_data = get_spline_derivatives_wrt_y(
            order, x_proc, y_proc, endpoint_derivatives_proc
        )

        derivative_arrays = derivative_data[:-1] if order == 3 else derivative_data

        x_min, x_max = x_proc[0], x_proc[-1]
        x_eval = np.linspace(x_min, x_max, num_points)

        # extract the gradient with respect to y values only
        g_y = g[1] if isinstance(g, tuple) else g
        if reversed_order:
            g_y = g_y[::-1]

        poly_factors, _ = accumulate_polynomial_factors(x_proc, x_eval, g_y, order)

        dv_dy = np.zeros(len(y_proc))
        n = len(x_proc) - 1
        for i in range(n):
            for k, derivative_array in enumerate(derivative_arrays):
                dv_dy += derivative_array[i, :] * poly_factors[i, k]
        if reversed_order:
            dv_dy = dv_dy[::-1]
        return dv_dy

    return vjp


defvjp(_interpolate_spline, None, interpolate_spline_y_vjp)


def interpolate_spline(
    x_points: NDArray,
    y_points: NDArray,
    num_points: int,
    order: int,
    endpoint_derivatives: tuple[Optional[float], Optional[float]] = (None, None),
) -> tuple[NDArray, NDArray]:
    """Differentiable spline interpolation of a given order
    with optional endpoint derivatives.

    Parameters
    ----------
    x_points : np.ndarray
        X coordinates of the data points (must be strictly monotonic)
    y_points : np.ndarray
        Y coordinates of the data points
    num_points : int
        Number of points in the output interpolation
    order : int
        Order of the spline (1=linear, 2=quadratic, 3=cubic)
    endpoint_derivatives : tuple[float, float] = (None, None)
        Derivatives at the endpoints (left, right)
        Note: For order=1 (linear), all endpoint derivatives are ignored.
              For order=2 (quadratic), only the left endpoint derivative is used.
              For order=3 (cubic), both endpoint derivatives are used if provided.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Tuple of (x_interpolated, y_interpolated) values

    Examples
    --------
    >>> import numpy as np
    >>> x = np.array([0, 1, 2])
    >>> y = np.array([0, 1, 0])
    >>> # Linear interpolation
    >>> x_interp, y_interp = interpolate_spline(x, y, num_points=5, order=1)
    >>> print(y_interp)
    [0.   0.5  1.   0.5  0. ]

    >>> # Quadratic interpolation with left endpoint derivative
    >>> x_interp, y_interp = interpolate_spline(x, y, num_points=5, endpoint_derivatives=(0, None), order=2)
    >>> print(np.round(y_interp, 3))
    [0.    0.25  1.    1.25   0.  ]

    >>> # Cubic interpolation with both endpoint derivatives
    >>> x_interp, y_interp = interpolate_spline(x, y, num_points=5, endpoint_derivatives=(0, 0), order=3)
    >>> print(np.round(y_interp, 3))
    [0.    0.5  1.    0.5  0.  ]
    """
    return _interpolate_spline(
        x_points,
        y_points,
        num_points,
        order,
        endpoint_derivatives,
    )



"""Defines time modulation to the medium"""

from __future__ import annotations

from abc import ABC, abstractmethod
from math import isclose
from typing import Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.constants import HERTZ, RADIAN
from tidy3d.exceptions import ValidationError

from .base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
from .data.data_array import SpatialDataArray
from .data.validators import validate_no_nans
from .time import AbstractTimeDependence
from .types import Bound, InterpMethod


class AbstractTimeModulation(AbstractTimeDependence, ABC):
    """Base class for modulation in time.

    Note
    ----
    This class describes the time dependence part of the separable space-time modulation type
    as shown below,

    .. math::

        amp(r, t) = \\Re[amp\\_time(t) \\cdot amp\\_space(r)]

    """

    @cached_property
    @abstractmethod
    def max_modulation(self) -> float:
        """Estimated maximum modulation amplitude."""


class ContinuousWaveTimeModulation(AbstractTimeDependence):
    """Class describing modulation with a harmonic time dependence.

    Note
    ----
    .. math::

        amp\\_time(t) = amplitude \\cdot \\
                e^{i \\cdot phase - 2 \\pi i \\cdot freq0 \\cdot t}

    Note
    ----
    The full space-time modulation is,

    .. math::

        amp(r, t) = \\Re[amp\\_time(t) \\cdot amp\\_space(r)]


    Example
    -------
    >>> cw = ContinuousWaveTimeModulation(freq0=200e12, amplitude=1, phase=0)
    """

    freq0: pd.PositiveFloat = pd.Field(
        ..., title="Modulation Frequency", description="Modulation frequency.", units=HERTZ
    )

    def amp_time(self, time: float) -> complex:
        """Complex-valued source amplitude as a function of time."""

        omega = 2 * np.pi * self.freq0
        return self.amplitude * np.exp(-1j * omega * time + 1j * self.phase)

    @cached_property
    def max_modulation(self) -> float:
        """Estimated maximum modulation amplitude."""
        return abs(self.amplitude)


TimeModulationType = Union[ContinuousWaveTimeModulation]


class AbstractSpaceModulation(ABC, Tidy3dBaseModel):
    """Base class for modulation in space.

    Note
    ----
    This class describes the 2nd term in the full space-time modulation below,
    .. math::

        amp(r, t) = \\Re[amp\\_time(t) \\cdot amp\\_space(r)]

    """

    @cached_property
    @abstractmethod
    def max_modulation(self) -> float:
        """Estimated maximum modulation amplitude."""


class SpaceModulation(AbstractSpaceModulation):
    """The modulation profile with a user-supplied spatial distribution of
    amplitude and phase.

    Note
    ----
    .. math::

        amp\\_space(r) = amplitude(r) \\cdot e^{i \\cdot phase(r)}

    The full space-time modulation is,

    .. math::

        amp(r, t) = \\Re[amp\\_time(t) \\cdot amp\\_space(r)]

    Example
    -------
    >>> Nx, Ny, Nz = 10, 9, 8
    >>> X = np.linspace(-1, 1, Nx)
    >>> Y = np.linspace(-1, 1, Ny)
    >>> Z = np.linspace(-1, 1, Nz)
    >>> coords = dict(x=X, y=Y, z=Z)
    >>> amp = SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=coords)
    >>> phase = SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=coords)
    >>> space = SpaceModulation(amplitude=amp, phase=phase)
    """

    amplitude: Union[float, SpatialDataArray] = pd.Field(
        1,
        title="Amplitude of modulation in space",
        description="Amplitude of modulation that can vary spatially. "
        "It takes the unit of whatever is being modulated.",
    )

    phase: Union[float, SpatialDataArray] = pd.Field(
        0,
        title="Phase of modulation in space",
        description="Phase of modulation that can vary spatially.",
        units=RADIAN,
    )

    interp_method: InterpMethod = pd.Field(
        "nearest",
        title="Interpolation method",
        description="Method of interpolation to use to obtain values at spatial locations on the Yee grids.",
    )

    _no_nans_amplitude = validate_no_nans("amplitude")
    _no_nans_phase = validate_no_nans("phase")

    @pd.validator("amplitude", always=True)
    def _real_amplitude(cls, val):
        """Assert that the amplitude is real."""
        if np.iscomplexobj(val):
            raise ValidationError("'amplitude' must be real.")
        return val

    @pd.validator("phase", always=True)
    def _real_phase(cls, val):
        """Assert that the phase is real."""
        if np.iscomplexobj(val):
            raise ValidationError("'phase' must be real.")
        return val

    @cached_property
    def max_modulation(self) -> float:
        """Estimated maximum modulation amplitude."""
        return np.max(abs(np.array(self.amplitude)))

    def sel_inside(self, bounds: Bound) -> SpaceModulation:
        """Return a new space modulation that contains the minimal amount data necessary to cover
        a spatial region defined by ``bounds``.


        Parameters
        ----------
        bounds : Tuple[float, float, float], Tuple[float, float float]
            Min and max bounds packaged as ``(minx, miny, minz), (maxx, maxy, maxz)``.

        Returns
        -------
        SpaceModulation
            SpaceModulation with reduced data.
        """

        if isinstance(self.amplitude, SpatialDataArray):
            amp_reduced = self.amplitude.sel_inside(bounds)
        else:
            amp_reduced = self.amplitude

        if isinstance(self.phase, SpatialDataArray):
            phase_reduced = self.phase.sel_inside(bounds)
        else:
            phase_reduced = self.phase

        return self.updated_copy(amplitude=amp_reduced, phase=phase_reduced)


SpaceModulationType = Union[SpaceModulation]


class SpaceTimeModulation(Tidy3dBaseModel):
    """Space-time modulation applied to a medium, adding
    on top of the time-independent part.


    Note
    ----
    The space-time modulation must be separable in space and time.
    e.g. when applied to permittivity,

    .. math::

        \\delta \\epsilon(r, t) = \\Re[amp\\_time(t) \\cdot amp\\_space(r)]
    """

    space_modulation: SpaceModulationType = pd.Field(
        SpaceModulation(),
        title="Space modulation",
        description="Space modulation part from the separable SpaceTimeModulation.",
        # discriminator=TYPE_TAG_STR,
    )

    time_modulation: TimeModulationType = pd.Field(
        ...,
        title="Time modulation",
        description="Time modulation part from the separable SpaceTimeModulation.",
        # discriminator=TYPE_TAG_STR,
    )

    @cached_property
    def max_modulation(self) -> float:
        """Estimated maximum modulation amplitude."""
        return self.time_modulation.max_modulation * self.space_modulation.max_modulation

    @cached_property
    def negligible_modulation(self) -> bool:
        """whether the modulation is weak enough to be regarded as zero."""
        # if isclose(np.diff(time_modulation.range), 0) and
        if isclose(self.max_modulation, 0):
            return True
        return False

    def sel_inside(self, bounds: Bound) -> SpaceTimeModulation:
        """Return a new space-time modulation that contains the minimal amount data necessary to cover
        a spatial region defined by ``bounds``.


        Parameters
        ----------
        bounds : Tuple[float, float, float], Tuple[float, float float]
            Min and max bounds packaged as ``(minx, miny, minz), (maxx, maxy, maxz)``.

        Returns
        -------
        SpaceTimeModulation
            SpaceTimeModulation with reduced data.
        """

        return self.updated_copy(space_modulation=self.space_modulation.sel_inside(bounds))


class ModulationSpec(Tidy3dBaseModel):
    """Specification adding space-time modulation to the non-dispersive part of medium
    including relative permittivity at infinite frequency and electric conductivity.
    """

    permittivity: SpaceTimeModulation = pd.Field(
        None,
        title="Space-time modulation of relative permittivity",
        description="Space-time modulation of relative permittivity at infinite frequency "
        "applied on top of the base permittivity at infinite frequency.",
    )

    conductivity: SpaceTimeModulation = pd.Field(
        None,
        title="Space-time modulation of conductivity",
        description="Space-time modulation of electric conductivity "
        "applied on top of the base conductivity.",
    )

    @pd.validator("conductivity", always=True)
    @skip_if_fields_missing(["permittivity"])
    def _same_modulation_frequency(cls, val, values):
        """Assert same time-modulation applied to permittivity and conductivity."""
        permittivity = values.get("permittivity")
        if val is not None and permittivity is not None:
            if val.time_modulation != permittivity.time_modulation:
                raise ValidationError(
                    "'permittivity' and 'conductivity' should have the same time modulation."
                )
        return val

    @cached_property
    def applied_modulation(self) -> bool:
        """Check if any modulation has been applied to ``permittivity`` or ``conductivity``."""
        return self.permittivity is not None or self.conductivity is not None

    def sel_inside(self, bounds: Bound) -> ModulationSpec:
        """Return a new modulation specficiation that contains the minimal amount data necessary to cover
        a spatial region defined by ``bounds``.


        Parameters
        ----------
        bounds : Tuple[float, float, float], Tuple[float, float float]
            Min and max bounds packaged as ``(minx, miny, minz), (maxx, maxy, maxz)``.

        Returns
        -------
        ModulationSpec
            ModulationSpec with reduced data.
        """

        perm_reduced = None
        if self.permittivity is not None:
            perm_reduced = self.permittivity.sel_inside(bounds)

        cond_reduced = None
        if self.conductivity is not None:
            cond_reduced = self.conductivity.sel_inside(bounds)

        return self.updated_copy(permittivity=perm_reduced, conductivity=cond_reduced)



from __future__ import annotations

import numpy as np

# default number of points per wvl in material for discretizing cylinder in autograd derivative
PTS_PER_WVL_MAT_CYLINDER_DISCRETIZE = 10

MAX_NUM_TRACED_STRUCTURES = 500
MAX_NUM_ADJOINT_PER_FWD = 10

GRADIENT_PRECISION = "single"  # Options: "single", "double"
GRADIENT_DTYPE_FLOAT = np.float32 if GRADIENT_PRECISION == "single" else np.float64
GRADIENT_DTYPE_COMPLEX = np.complex64 if GRADIENT_PRECISION == "single" else np.complex128

GAUSS_QUADRATURE_ORDER = 7
QUAD_SAMPLE_FRACTION = 0.4

AUTOGRAD_MONITOR_INTERVAL_SPACE_POLY = (1, 1, 1)
AUTOGRAD_MONITOR_INTERVAL_SPACE_CUSTOM = (1, 1, 1)

DEFAULT_WAVELENGTH_FRACTION = 0.1
MINIMUM_SPACING = 1e-2

EDGE_CLIP_TOLERANCE = 1e-9

# chunk size for processing multiple frequencies in adjoint gradient computation.
# None = process all frequencies at once (no chunking)
ADJOINT_FREQ_CHUNK_SIZE = None



"""Class and custom data array for representing a scattering matrix wave port."""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property, skip_if_fields_missing
from tidy3d.components.data.data_array import FreqDataArray, FreqModeDataArray
from tidy3d.components.data.monitor_data import ModeData
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.geometry.base import Box
from tidy3d.components.geometry.bound_ops import bounds_contains
from tidy3d.components.grid.grid import Grid
from tidy3d.components.monitor import ModeMonitor
from tidy3d.components.simulation import Simulation
from tidy3d.components.source.field import ModeSource, ModeSpec
from tidy3d.components.source.time import GaussianPulse
from tidy3d.components.structure import MeshOverrideStructure
from tidy3d.components.types import Axis, Direction, FreqArray
from tidy3d.constants import fp_eps
from tidy3d.exceptions import ValidationError
from tidy3d.plugins.microwave import CurrentIntegralTypes, ImpedanceCalculator, VoltageIntegralTypes
from tidy3d.plugins.mode import ModeSolver

from .base_terminal import AbstractTerminalPort

DEFAULT_WAVE_PORT_NUM_CELLS = 5
MIN_WAVE_PORT_NUM_CELLS = 3


class WavePort(AbstractTerminalPort, Box):
    """Class representing a single wave port"""

    direction: Direction = pd.Field(
        ...,
        title="Direction",
        description="'+' or '-', defining which direction is considered 'input'.",
    )

    mode_spec: ModeSpec = pd.Field(
        ModeSpec(),
        title="Mode Specification",
        description="Parameters to feed to mode solver which determine modes measured by monitor.",
    )

    mode_index: pd.NonNegativeInt = pd.Field(
        0,
        title="Mode Index",
        description="Index into the collection of modes returned by mode solver. "
        " Specifies which mode to inject using this source. "
        "If larger than ``mode_spec.num_modes``, "
        "``num_modes`` in the solver will be set to ``mode_index + 1``.",
    )

    voltage_integral: Optional[VoltageIntegralTypes] = pd.Field(
        None,
        title="Voltage Integral",
        description="Definition of voltage integral used to compute voltage and the characteristic impedance.",
    )

    current_integral: Optional[CurrentIntegralTypes] = pd.Field(
        None,
        title="Current Integral",
        description="Definition of current integral used to compute current and the characteristic impedance.",
    )

    num_grid_cells: Optional[int] = pd.Field(
        DEFAULT_WAVE_PORT_NUM_CELLS,
        ge=MIN_WAVE_PORT_NUM_CELLS,
        title="Number of Grid Cells",
        description="Number of mesh grid cells in the transverse plane of the `WavePort`. "
        "Used in generating the suggested list of :class:`.MeshOverrideStructure` objects. "
        "Must be greater than or equal to 3. When set to `None`, no grid refinement is performed.",
    )

    def _mode_voltage_coefficients(self, mode_data: ModeData) -> FreqModeDataArray:
        """Calculates scaling coefficients to convert mode amplitudes
        to the total port voltage.
        """
        mode_data = mode_data._isel(mode_index=[self.mode_index])
        if self.voltage_integral is None:
            current_coeffs = self.current_integral.compute_current(mode_data)
            voltage_coeffs = 2 * np.abs(mode_data.flux) / np.conj(current_coeffs)
        else:
            voltage_coeffs = self.voltage_integral.compute_voltage(mode_data)
        return voltage_coeffs.squeeze()

    def _mode_current_coefficients(self, mode_data: ModeData) -> FreqModeDataArray:
        """Calculates scaling coefficients to convert mode amplitudes
        to the total port current.
        """
        mode_data = mode_data._isel(mode_index=[self.mode_index])
        if self.current_integral is None:
            voltage_coeffs = self.voltage_integral.compute_voltage(mode_data)
            current_coeffs = (2 * np.abs(mode_data.flux) / voltage_coeffs).conj()
        else:
            current_coeffs = self.current_integral.compute_current(mode_data)
        return current_coeffs.squeeze()

    @cached_property
    def injection_axis(self) -> Axis:
        """Injection axis of the port."""
        return self.size.index(0.0)

    @cached_property
    def transverse_axes(self) -> tuple[Axis, Axis]:
        """Transverse axes of the port."""
        _, trans_axes = Box.pop_axis([0, 1, 2], self.injection_axis)
        return trans_axes

    @cached_property
    def _mode_monitor_name(self) -> str:
        """Return the name of the :class:`.ModeMonitor` associated with this port."""
        return f"{self.name}_mode"

    def to_source(
        self, source_time: GaussianPulse, snap_center: Optional[float] = None
    ) -> ModeSource:
        """Create a mode source from the wave port."""
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center
        return ModeSource(
            center=center,
            size=self.size,
            source_time=source_time,
            mode_spec=self.mode_spec,
            mode_index=self.mode_index,
            direction=self.direction,
            name=self.name,
        )

    def to_monitors(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> list[ModeMonitor]:
        """The wave port uses a :class:`.ModeMonitor` to compute the characteristic impedance
        and the port voltages and currents."""
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center
        mode_mon = ModeMonitor(
            center=self.center,
            size=self.size,
            freqs=freqs,
            name=self._mode_monitor_name,
            colocate=False,
            mode_spec=self.mode_spec,
            store_fields_direction=self.direction,
        )
        return [mode_mon]

    def to_mode_solver(self, simulation: Simulation, freqs: FreqArray) -> ModeSolver:
        """Helper to create a :class:`.ModeSolver` instance."""
        mode_solver = ModeSolver(
            simulation=simulation,
            plane=self.geometry,
            mode_spec=self.mode_spec,
            freqs=freqs,
            direction=self.direction,
            colocate=False,
        )
        return mode_solver

    def compute_voltage(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute voltage across the port."""
        mode_data = sim_data[self._mode_monitor_name]
        voltage_coeffs = self._mode_voltage_coefficients(mode_data)
        amps = mode_data.amps
        fwd_amps = amps.sel(direction="+").squeeze()
        bwd_amps = amps.sel(direction="-").squeeze()
        return voltage_coeffs * (fwd_amps + bwd_amps)

    def compute_current(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute current flowing through the port."""
        mode_data = sim_data[self._mode_monitor_name]
        current_coeffs = self._mode_current_coefficients(mode_data)
        amps = mode_data.amps
        fwd_amps = amps.sel(direction="+").squeeze()
        bwd_amps = amps.sel(direction="-").squeeze()
        # In ModeData, fwd_amps and bwd_amps are not relative to
        # the direction fields are stored
        sign = 1.0
        if self.direction == "-":
            sign = -1.0
        return sign * current_coeffs * (fwd_amps - bwd_amps)

    def compute_port_impedance(
        self, sim_mode_data: Union[SimulationData, ModeData]
    ) -> FreqModeDataArray:
        """Helper to compute impedance of port. The port impedance is computed from the
        transmission line mode, which should be TEM or at least quasi-TEM."""
        impedance_calc = ImpedanceCalculator(
            voltage_integral=self.voltage_integral, current_integral=self.current_integral
        )
        if isinstance(sim_mode_data, SimulationData):
            mode_data = sim_mode_data[self._mode_monitor_name]
        else:
            mode_data = sim_mode_data

        # Filter out unwanted modes to reduce impedance computation effort
        mode_data = mode_data._isel(mode_index=[self.mode_index])
        impedance_array = impedance_calc.compute_impedance(mode_data)
        return impedance_array

    def to_mesh_overrides(self) -> list[MeshOverrideStructure]:
        """Creates a list of :class:`.MeshOverrideStructure` for mesh refinement in the transverse
        plane of the port. The mode source requires at least 3 grid cells in the transverse
        dimensions, so these mesh overrides will be added to the simulation to ensure that this
        requirement is satisfied.
        """
        dl = [None] * 3
        for trans_axis in self.transverse_axes:
            dl[trans_axis] = self.size[trans_axis] / self.num_grid_cells

        return [
            MeshOverrideStructure(
                geometry=Box(center=self.center, size=self.size),
                dl=dl,
                shadow=False,
                priority=-1,
            )
        ]

    @pd.validator("voltage_integral", "current_integral")
    def _validate_path_integrals_within_port(cls, val, values):
        """Raise ``ValidationError`` when the supplied path integrals are not within the port bounds."""
        center = values["center"]
        size = values["size"]
        box = Box(center=center, size=size)
        if val and not bounds_contains(
            box.bounds, val.bounds, fp_eps, np.finfo(np.float32).smallest_normal
        ):
            raise ValidationError(
                f"'{cls.__name__}' must be setup with all path integrals defined within the bounds "
                f"of the port. Path bounds are '{val.bounds}', but port bounds are '{box.bounds}'."
            )
        return val

    @pd.validator("current_integral", always=True)
    @skip_if_fields_missing(["voltage_integral"])
    def _check_voltage_or_current(cls, val, values):
        """Raise validation error if both ``voltage_integral`` and ``current_integral``
        were not provided."""
        if values.get("voltage_integral") is None and val is None:
            raise ValidationError(
                "At least one of 'voltage_integral' or 'current_integral' must be provided."
            )
        return val

    @pd.validator("current_integral", always=True)
    def _validate_current_integral_sign(cls, val, values):
        """
        Validate that the sign of ``current_integral`` matches the port direction.
        """
        if val is None:
            return val

        direction = values.get("direction")
        name = values.get("name")
        if val.sign != direction:
            raise ValidationError(
                f"'current_integral' sign must match the '{name}' direction '{direction}'."
            )
        return val



"""Computes approximate parameters related to microstrip transmission lines using the models
in [1,2].

References
----------
[1]     E. Hammerstad and O. Jensen, "Accurate Models for Microstrip Computer-Aided Design,"
         1980 IEEE MTT-S International Microwave Symposium Digest,
         Washington, DC, USA, 1980, pp. 407-409.

[2]     Kirschning, Manfred Klaus, Rolf H. Jansen and Norbert H. L. Koster. “Accurate model
        for open end effect of microstrip lines.” Electronics Letters 17 (1981): 123-125.
"""

from __future__ import annotations

import numpy as np

from tidy3d.constants import ETA_0


def _f(normalized_width: float) -> float:
    """Equation 2 [1]"""
    u = normalized_width
    return 6 + (2 * np.pi - 6) * np.exp(-((30.666 / u) ** 0.7528))


def _z01(normalized_width: float) -> float:
    """Equation 1 [1]"""
    u = normalized_width
    f_u = _f(u)
    return ETA_0 / (2 * np.pi) * np.log(f_u / u + np.sqrt(1 + (2 / u) ** 2))


def _a(normalized_width: float) -> float:
    """Equation 4 [1]"""
    u = normalized_width
    u4 = u**4
    second_term = (1 / 49) * np.log((u4 + (u / 52) ** 2) / (u4 + 0.432))
    third_term = (1 / 18.7) * np.log(1 + (u / 18.1) ** 3)
    return 1 + second_term + third_term


def _b(relative_permittivity: float) -> float:
    """Equation 5 [1]"""
    er = relative_permittivity
    return 0.564 * ((er - 0.9) / (er + 3)) ** (0.053)


def _epsilon_e(normalized_width: float, relative_permittivity: float) -> float:
    """Equation 3 [1]"""
    u = normalized_width
    er = relative_permittivity
    a_u = _a(u)
    b_er = _b(er)
    scale_factor = (1 + 10 / u) ** (-a_u * b_er)
    return (er + 1) / 2 + (er - 1) / 2 * scale_factor


def _wcorr_homo(normalized_width: float, strip_thickness: float) -> float:
    """Equation 6 [1]"""
    u = normalized_width
    t = strip_thickness
    coth = 1 / np.tanh(np.sqrt(6.517 * u))
    quotient = 4 * np.exp(1) / (t * (coth) ** 2)
    return t / np.pi * np.log(1 + quotient)


def _wcorr_mixed(width_correction_homo: float, relative_permittivity: float) -> float:
    """Equation 7 [1]"""
    wch = width_correction_homo
    er = relative_permittivity
    return 0.5 * (1 + 1 / np.cosh(np.sqrt(er - 1))) * wch


def compute_line_params(
    relative_permittivity: float, width: float, height: float, thickness: float
) -> tuple[float, float]:
    """Computes an approximation for the characteristic impedance and effective
    electric permittivity of a microstrip line [1].

    Parameters
    ----------
    relative_permittivity :
        Relative electric permittivity of the substrate material.
    width :
        Width of metal strip.
    height :
        Distance between ground and bottom of strip.
    thickness :
        Thickness of metal strip.

    Returns
    -------
    ``tuple``
        ``Tuple`` containing the characteristic impedance and effective relative permittivity.
    """

    normalized_width = width / height
    normalized_thickness = thickness / height
    u_1 = normalized_width
    u_r = normalized_width
    t = normalized_thickness
    # Add correction to width if nonzero strip thickness is given
    if thickness is not None and thickness != 0:
        delta_wcorr_homo = _wcorr_homo(normalized_width, t)
        delta_wcorr = _wcorr_mixed(delta_wcorr_homo, relative_permittivity)
        u_1 += delta_wcorr_homo
        u_r += delta_wcorr
    Z_01_r = _z01(u_r)
    Z_01_1 = _z01(u_1)
    er_e = _epsilon_e(u_r, relative_permittivity)
    Z_0 = Z_01_r / np.sqrt(er_e)
    er_eff = er_e * (Z_01_1 / Z_01_r) ** 2
    return (Z_0, er_eff)


def compute_end_effect_length(
    relative_permittivity: float, er_eff: float, width: float, height: float
) -> float:
    """Computes the effective increase in length of a microstrip line which has been terminated by an
    open-circuit. Equations 1,2 [2].

    Parameters
    ----------
    relative_permittivity :
        Relative electric permittivity of the substrate material.
    er_eff :
        Effective relative electric permittivity of the microstrip.
    width :
        Width of strip.
    height :
        Distance between ground and bottom of strip.

    Returns
    -------
    float
        The effective additional length of the microstrip."""

    normalized_width = width / height
    u = normalized_width
    er = relative_permittivity
    Xi1 = (
        0.434907
        * (er_eff**0.81 + 0.26 * u**0.8544 + 0.236)
        / (er_eff**0.81 - 0.189 * u**0.8544 + 0.87)
    )
    Xi2 = 1 + u**0.371 / (2.358 * er + 1)
    Xi3 = 1 + (0.5274 * np.arctan(0.084 * u ** (1.9413 / Xi2))) / (er_eff**0.9236)
    Xi4 = 1 + 0.0377 * np.arctan(0.067 * u**1.456) * (6 - 5 * np.exp(0.036 * (1 - er)))
    Xi5 = 1 - 0.218 * np.exp(-7.5 * u)
    delta_L = height * Xi1 * Xi3 * Xi5 / Xi4
    return delta_L



"""Defines information about a task"""

from __future__ import annotations

from abc import ABC
from datetime import datetime
from enum import Enum
from typing import Optional

import pydantic.v1 as pydantic


class TaskStatus(Enum):
    """The statuses that the task can be in."""

    INIT = "initialized"
    """The task has been initialized."""

    QUEUE = "queued"
    """The task is in the queue."""

    PRE = "preprocessing"
    """The task is in the preprocessing stage."""

    RUN = "running"
    """The task is running."""

    POST = "postprocessing"
    """The task is in the postprocessing stage."""

    SUCCESS = "success"
    """The task has completed successfully."""

    ERROR = "error"
    """The task has completed with an error."""


class TaskBase(pydantic.BaseModel, ABC):
    """Base configuration for all task objects."""

    class Config:
        """Configuration for TaskBase"""

        arbitrary_types_allowed = True
        """Allow arbitrary types to be used within the model."""


class ChargeType(str, Enum):
    """The payment method of the task."""

    FREE = "free"
    """No payment required."""

    PAID = "paid"
    """Payment required."""


class TaskBlockInfo(TaskBase):
    """Information about the task's block status.

    This includes details about how the task can be blocked by various features
    such as user limits and insufficient balance.
    """

    chargeType: ChargeType = None
    """The type of charge applicable to the task (free or paid)."""

    maxFreeCount: int = None
    """The maximum number of free tasks allowed."""

    maxGridPoints: int = None
    """The maximum number of grid points permitted."""

    maxTimeSteps: int = None
    """The maximum number of time steps allowed."""


class TaskInfo(TaskBase):
    """General information about a task."""

    taskId: str
    """Unique identifier for the task."""

    taskName: str = None
    """Name of the task."""

    nodeSize: int = None
    """Size of the node allocated for the task."""

    completedAt: Optional[datetime] = None
    """Timestamp when the task was completed."""

    status: str = None
    """Current status of the task."""

    realCost: float = None
    """Actual cost incurred by the task."""

    timeSteps: int = None
    """Number of time steps involved in the task."""

    solverVersion: str = None
    """Version of the solver used for the task."""

    createAt: Optional[datetime] = None
    """Timestamp when the task was created."""

    estCostMin: float = None
    """Estimated minimum cost for the task."""

    estCostMax: float = None
    """Estimated maximum cost for the task."""

    realFlexUnit: float = None
    """Actual flexible units used by the task."""

    oriRealFlexUnit: float = None
    """Original real flexible units."""

    estFlexUnit: float = None
    """Estimated flexible units for the task."""

    estFlexCreditTimeStepping: float = None
    """Estimated flexible credits for time stepping."""

    estFlexCreditPostProcess: float = None
    """Estimated flexible credits for post-processing."""

    estFlexCreditMode: float = None
    """Estimated flexible credits based on the mode."""

    s3Storage: float = None
    """Amount of S3 storage used by the task."""

    startSolverTime: Optional[datetime] = None
    """Timestamp when the solver started."""

    finishSolverTime: Optional[datetime] = None
    """Timestamp when the solver finished."""

    totalSolverTime: int = None
    """Total time taken by the solver."""

    callbackUrl: str = None
    """Callback URL for task notifications."""

    taskType: str = None
    """Type of the task."""

    metadataStatus: str = None
    """Status of the metadata for the task."""

    taskBlockInfo: TaskBlockInfo = None
    """Blocking information for the task."""


class RunInfo(TaskBase):
    """Information about the run of a task."""

    perc_done: pydantic.confloat(ge=0.0, le=100.0)
    """Percentage of the task that is completed (0 to 100)."""

    field_decay: pydantic.confloat(ge=0.0, le=1.0)
    """Field decay from the maximum value (0 to 1)."""

    def display(self):
        """Print some info about the task's progress."""
        print(f" - {self.perc_done:.2f} (%) done")
        print(f" - {self.field_decay:.2e} field decay from max")
